open Util
open Combinator

(** FPAT Configuration *)

let _ =
  HCCSSolver.ext_of_string :=
    (function
      | "git" -> GenHCCSSolver.solve (CHGenInterpProver.interpolate false)
      | "gch" -> GenHCCSSolver.solve (CHGenInterpProver.interpolate true)
      | _ -> raise Not_found)

let usage =
  (*"FPAT: A Framework for Program Analysis and Transformation\n\n" ^*)
  "Usage: "
  ^ Sys.executable_name
  ^ " <options> <files>\noptions are:"
let arg_spec =
  ["-silent", Arg.Set Global.silent, " Enable silent mode";
   "-bench", Arg.Set Global.bench_mode, " Enable benchmark mode";
   "-timeout", Arg.Int (fun n -> Global.timeout := n), "<n> Set timeout (sec)";
   (* debugging *)
   "-debug", Arg.Set Global.debug, " Enable debug mode";
   "-debug-regexp", Arg.String (fun str -> Global.debug := true;
                                           Logger.filter_regexp := Str.regexp str),
   "<regexp> Enable debug mode (only log calls that match <regexp> and their ancestor calls)";
   "-debug-regexp-exact", Arg.String (fun str -> Global.debug := true;
                                                 Logger.filter_exact := true;
                                                 Logger.filter_regexp := Str.regexp str),
   "<regexp> Enable debug mode (only log calls that match <regexp>)";
   "-debug-call-id", Arg.Int (fun n -> Global.debug := true;
                                       Logger.filter_call_id := n),
   "<n> Enable debug mode (only log calls that match call_id <n> and their ancestor calls)";
   "-debug-call-id-exact", Arg.Int (fun n -> Global.debug := true;
                                             Logger.filter_exact := true;
                                             Logger.filter_call_id := n),
   "<n> Enable debug mode (only log calls that match call_id <n>)";
   "-debug-level", Arg.Int (fun n -> Global.debug_level := n),
   "<n> Set debug level";
   "-profile", Arg.Unit (fun () -> Global.debug := true;
                                   Logger.filter_kinds := [Logger.Trace]),
   " Enable profiling mode";
   "-print-log", Arg.Set Global.print_log,
   " Print log messages to the standard output";
   "-disable-print-callstack-exception",
   Arg.Clear Logger.print_call_stack_on_exception,
   " Disable printing the call stack when an exception is occurred";
   (* options for program verification *)
   (* preprocessing of ML programs *)
   "-no-anf", Arg.Clear Fdef.use_anf, " Disable ANF transformation";
   (* HCCS solvers *)
   "-hccs", Arg.String (fun str ->
                try HCCSSolver.of_string_dyn str |> HCCSSolver.link_dyn
                with Not_found -> raise (Arg.Bad "invalid argument for -hccs")),
   "<n> Use an HCCS solver based on\n"
   ^ "\t\t\t\tgit: generalized interpolation\n"
   ^ "\t\t\t\tgch: generalized interpolation using convex hull\n";
   (* predicate abstraction *)
   "-wp-max", Arg.Set_int PredAbst.wp_max_num,
   "<n> The maximum dimension of hypercubes in predicate abstraction";
   "-neg-pred", Arg.Set PredAbst.use_neg_pred,
   " Use negative predicates for abstraction";
  ]



let set_default () =
  (* default SMT solver *)
  SMTProver.init_z3 ();
  (* default polynomial constraint solver *)
  PolyConstrSolver.ext_solve := PolyConstrSolver.solve_z3;
  (* default interpolating prover *)
  InterpProver.link_dyn InterpProver.interpolate_csisat_dyn;
  (* default HCCS solver *)
  HCCSSolver.link_dyn BwIPHCCSSolver.solve;
  (* default unit, bool, and tuple HCCS encoders *)
  HCCSSolver.ext_solve_unit := UnitHCCSSolver.solve;
  HCCSSolver.ext_solve_bool := id;
  HCCSSolver.ext_solve_tuple := TupHCCSSolver.solve
let set_default =
  (* this is not printed because !Global.debug is false *)
  Logger.log_block1 "FPATConfig.set_default" set_default

let is_fpat_exception = function
  | Global.NotImplemented _
  | Global.NoMatch _
  (*| SMTProver.Unsat*)
  | SMTProver.Unknown
  (*| InterpProver.NoInterpolant*)
  | InterpProver.Fail
  | InterpProver.Unknown
  (*| PolyConstrSolver.NoSolution*)
  | PolyConstrSolver.Unknown
  | RefTypInfer.FailedToRefineTypes
  | RefTypInfer.FailedToRefineInputTypes
  | RefTypInfer.FailedToRefineExtraParameters -> true
  | _ -> false

let pr_exception ppf = function
  | Global.NotImplemented msg -> Format.fprintf ppf "Not implemented: %s" msg
  | Global.NoMatch msg -> Format.fprintf ppf "Not matched: %s" msg
  | SMTProver.Unknown -> Format.fprintf ppf "Failure of SMT prover"
  | InterpProver.Fail ->
    Format.fprintf ppf
      "Failure of interpolating prover (integer domain not fully supported)"
  | InterpProver.Unknown ->
    Format.fprintf ppf "Failure of interpolating prover"
  | PolyConstrSolver.Unknown ->
    Format.fprintf ppf "Failure of polynomial constraint solver"
  | RefTypInfer.FailedToRefineTypes ->
    Format.fprintf ppf "Failure of abstraction type refinement"
  | RefTypInfer.FailedToRefineInputTypes ->
    Format.fprintf ppf
      "Failure of abstraction type refinement of external inputs"
  | RefTypInfer.FailedToRefineExtraParameters ->
    Format.fprintf ppf "Failure of parameter substitution refinement"
  | _ -> raise Not_found

let string_of_fpat_exception = Printer.string_of pr_exception
