open Util
open CEGAR_syntax
open CEGAR_type
open CEGAR_util
open CEGAR_trans


exception CannotRefute

let add_preds_env map env =
  let aux (f,typ) =
    let typ' =
      try
        merge_typ typ @@ List.assoc f map
      with Not_found -> typ
    in
    f, typ'
  in
  List.map aux env

let add_renv map env =
  let aux (n, preds) = make_randint_name n, TBase(TInt, preds) in
  add_preds_env (List.map aux map) env

let new_id' x = new_id (Format.sprintf "%s_%d" x !Flag.cegar_loop)

let rec negate_typ = function
  | TBase(b,ps) ->
      let x = new_id' "x" in
      let ps t = List.map (make_not |- subst x t) (ps (Var x)) in
      TBase(b, ps)
  | TFun(typ1,typ2) ->
      let x = new_id' "x" in
      let typ2 = typ2 (Var x) in
      let typ1 = negate_typ typ1 in
      let typ2 = negate_typ typ2 in
      TFun(typ1, fun t -> subst_typ x t typ2)
  | (TAbs _ | TApp _) as typ -> Format.printf "negate_typ: %a." CEGAR_print.typ typ; assert false

let add_nag_preds_renv env =
  let aux (f,typ) = if is_randint_var f then merge_typ typ (negate_typ typ) else typ in
  List.map aux env

let add_preds map prog =
  {prog with env = add_preds_env map prog.env}


let rec add_to_path path typ1 typ2 =
  match path,typ2 with
  | [],_ -> merge_typ typ1 typ2
  | 0::path',TFun(typ21,typ22) -> TFun(add_to_path path' typ1 typ21, typ22)
  | 1::path',TFun(typ21,typ22) -> TFun(typ21, fun x -> add_to_path path' typ1 (typ22 x))
  | _ -> Format.printf "%a@." CEGAR_print.typ typ2; assert false

let rec add_pred n path typ =
  match typ with
  | TBase _ -> assert false
  | TFun(typ1,typ2) when n=0 ->
      TFun(typ1, fun x -> add_to_path (List.tl path) typ1 (typ2 x))
  | TFun(typ1,typ2) ->
      assert (List.hd path = 1);
      TFun(typ1, fun x -> add_pred (n-1) (List.tl path) (typ2 x))
  | TAbs _ -> assert false
  | TApp _ -> assert false



let refine labeled is_cp prefix ces ext_ces prog =
  let tmp = Time.get () in
  let post () =
    Fpat.SMTProver.finalize ();
    Fpat.SMTProver.initialize ();
    Time.add tmp Flag.time_cegar
  in
  try
    if !Flag.print_progress then
      Color.printf
        Color.Green
        "(%d-4) Discovering predicates (infeasible case) ... @."
        !Flag.cegar_loop;
    if Flag.use_prefix_trace then
      fatal "Not implemented: Flag.use_prefix_trace";
    let map =
      let ces,ext_ces =
        if !Flag.use_multiple_paths then
          ces, ext_ces
        else
          [List.hd ces], [List.hd ext_ces]
      in
      Format.printf "@[<v>";
      let map = FpatInterface.infer labeled is_cp ces ext_ces prog in
      Format.printf "@]";
      map
    in
    let env =
      if !Flag.disable_predicate_accumulation then
        map
      else
        add_preds_env map prog.env
    in
    if !Flag.print_progress then Format.printf "DONE!@.@.";
    post ();
    map, {prog with env}
  with e ->
    post ();
    raise e

let refine_with_ext labeled is_cp prefix ces ext_ces prog =
  let tmp = Time.get () in
  try
    if !Flag.print_progress then
      Color.printf
        Color.Green
        "(%d-4) Discovering predicates (feasible case) ... @."
        !Flag.cegar_loop;
    if Flag.use_prefix_trace then
      raise (Fatal "Not implemented: Flag.use_prefix_trace");
    Format.printf "@[<v>";
    let map = FpatInterface.infer_with_ext labeled is_cp ces ext_ces prog in
    Format.printf "@]";
    let env =
      if !Flag.disable_predicate_accumulation then
        map
      else
        add_preds_env map prog.env
    in
    if !Flag.print_progress then Format.printf "DONE!@.@.";
    Fpat.SMTProver.finalize ();
    Fpat.SMTProver.initialize ();
    Time.add tmp Flag.time_cegar;
    map, {prog with env}
  with e ->
    Fpat.SMTProver.finalize ();
    Fpat.SMTProver.initialize ();
    Time.add tmp Flag.time_cegar;
    raise e

exception PostCondition of (Fpat.Idnt.t * Fpat.Type.t) list * Fpat.Formula.t * Fpat.Formula.t

let print_list fm = function
  | [] -> Format.fprintf fm "[]@."
  | x::xs ->
    let rec iter = function
      | [] -> ""
      | y::ys -> ", " ^ string_of_int y ^ iter ys
    in
    Format.fprintf fm "[%d%s]@." x (iter xs)

let refine_rank_fun ce ex_ce prog =
  let tmp = Time.get () in
  try
    (*Format.printf "(%d)[refine_rank_fun] %a @." !Flag.cegar_loop print_list ce;
      Format.printf "    %a@." (print_prog_typ' [] []) { env=env; defs=defs; main=main };*)
    if !Flag.print_progress then Format.printf "(%d-4) Discovering ranking function ... @." !Flag.cegar_loop;
    let env, spc =
      Format.printf "@[<v>";
      let env, spc = FpatInterface.compute_strongest_post prog ce ex_ce in
      Format.printf "@]";
      env, spc
    in

    let spcWithExparam =
      if !Flag.add_closure_exparam
      then
        let progWithExparam = Option.get prog.info.exparam_orig in
        if false then Format.printf "REFINE: %a@." CEGAR_print.prog @@ Option.get prog.info.exparam_orig;
        Format.printf "@[<v>";
        let _, spcWithExparam = FpatInterface.compute_strongest_post progWithExparam ce ex_ce in
        Format.printf "@]";
        if false then Format.printf "REFINE: %a@." Fpat.Formula.pr spcWithExparam;
        spcWithExparam
      else spc (* dummy *)
    in

    (* TEMPORARY *)
    (*Format.printf "[exparam]@.%a@." FpatInterface.Formula.pr spcWithExparam;
      Format.printf "[instantiated]@.%a@." FpatInterface.Formula.pr spc;*)

    raise (PostCondition (env, spc, spcWithExparam))
  with e ->
    Fpat.SMTProver.finalize ();
    Fpat.SMTProver.initialize ();
    Time.add tmp Flag.time_cegar;
    raise e
